import os
from pathlib import Path

import graphene

from whitenoise import WhiteNoise
from wsgiadapter import WSGIAdapter as RequestsWSGIAdapter
from requests import Session as RequestsSession

from . import models
from .status import HTTP_404


class BaseAPI:
    __slots__ = ["routes"]

    def __init__(self):
        self.routes = {}

    def _wsgi_app(self, environ, start_response):
        # def wsgi_app(self, request):
        """The actual WSGI application. This is not implemented in
        :meth:`__call__` so that middlewares can be applied without
        losing a reference to the app object. Instead of doing this::

            app = MyMiddleware(app)

        It's a better idea to do this instead::

            app.wsgi_app = MyMiddleware(app.wsgi_app)

        Then you still have the original application object around and
        can continue to call methods on it.

        .. versionchanged:: 0.7
            Teardown events for the request and app contexts are called
            even if an unhandled error occurs. Other events may not be
            called depending on when an error occurs during dispatch.
            See :ref:`callbacks-and-errors`.

        :param environ: A WSGI environment.
        :param start_response: A callable accepting a status code,
            a list of headers, and an optional exception context to
            start the response.
        """

        req = models.Request.from_environ(environ)
        resp = self._dispatch_request(req)

        return resp(environ, start_response)

    def wsgi_app(self, environ, start_response):
        return self.whitenoise(environ, start_response)

    def __call__(self, environ, start_response):
        """The WSGI server calls the Flask application object as the
        WSGI application. This calls :meth:`wsgi_app` which can be
        wrapped to applying middleware."""
        return self.wsgi_app(environ, start_response)

    def path_matches_route(self, url):
        for (route, view) in self.routes.items():
            if url == route:
                return route

    def _dispatch_request(self, req):
        route = self.path_matches_route(req.path)
        resp = models.Response(req=req)

        if route:
            try:
                self.routes[route](req, resp)
            # The request is using class-based views.
            except TypeError:
                try:
                    view = self.routes[route]()
                # GraphQL Schema.
                except TypeError:
                    view = self.routes[route]
                    self.graphql_response(req, resp, schema=view)

                # Run on_request first.
                try:
                    getattr(view, "on_request")(req, resp)
                except AttributeError:
                    pass

                # Then on_get.
                method = req.method.lower()

                try:
                    getattr(view, f"on_{method}")(req, resp)
                except AttributeError:
                    pass
        else:
            self.default_response(req, resp)

        return resp

    @property
    def static_dir(self):
        return Path(".")


class API(BaseAPI):
    __slots__ = ("routes", "_session", "whitenoise", "static_dir")

    def __init__(self, static="static"):
        super().__init__()
        self._session = None
        self.static_dir = Path(os.path.abspath(static))

        # Make the static directory if it doesn't exist.
        os.makedirs(self.static_dir, exist_ok=True)

        # Mount the whitenoise application.
        self.whitenoise = WhiteNoise(self._wsgi_app, root=str(self.static_dir))

    def add_route(self, route, view, *, check_existing=True, graphiql=False):
        if check_existing:
            assert route not in self.routes

        # TODO: Support grpahiql.

        self.routes[route] = view

    def default_response(self, req, resp):
        resp.status_code = HTTP_404
        resp.text = "Not found."

    @staticmethod
    def _resolve_graphql_query(req):
        # Support query/q in form data.
        if "query" in req.data:
            return req.data["query"]
        if "q" in req.data:
            return req.data["q"]

        # Support query/q in params.
        if "query" in req.params:
            return req.params["query"][0]
        if "q" in req.params:
            return req.parama["q"][0]

        # Otherwise, the request text is used (typical).
        # TODO: Make some assertions about content-type here.
        return req.text

    def graphql_response(self, req, resp, schema):
        query = self._resolve_graphql_query(req)
        result = schema.execute(query)
        resp.media = dict(result.data)

    def route(self, route, **options):
        def decorator(f):
            self.add_route(route, f)
            return f

        return decorator

    def session(self, base_url="http://app"):
        if self._session is None:
            session = RequestsSession()
            session.mount(base_url, RequestsWSGIAdapter(self))
            self._session = session
        return self._session



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import codecs
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = "\n" + f.read()

about = {}

with open(os.path.join(here, "responder", "__version__.py")) as f:
    exec(f.read(), about)

if sys.argv[-1] == "publish":
    os.system("python setup.py sdist bdist_wheel upload")
    sys.exit()

required = []


# https://pypi.python.org/pypi/stdeb/0.8.5#quickstart-2-just-tell-me-the-fastest-way-to-make-a-deb
class DebCommand(Command):
    """Support for setup.py deb"""

    description = "Build and publish the .deb package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "deb_dist"))
        except FileNotFoundError:
            pass
        self.status(u"Creating debian mainfest…")
        os.system(
            "python setup.py --command-packages=stdeb.command sdist_dsc -z artful --package3=pipenv --depends3=python3-virtualenv-clone"
        )
        self.status(u"Building .deb…")
        os.chdir("deb_dist/pipenv-{0}".format(about["__version__"]))
        os.system("dpkg-buildpackage -rfakeroot -uc -us")


class UploadCommand(Command):
    """Support setup.py publish."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except FileNotFoundError:
            pass
        self.status("Building Source distribution…")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))
        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")
        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")
        sys.exit()


setup(
    name="responder",
    version=about["__version__"],
    description="A sorta familiar HTTP framework.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Kenneth Reitz",
    author_email="me@kennethreitz.org",
    url="https://github.com/kennethreitz/responder",
    packages=find_packages(exclude=["tests", "tests.*", "tasks", "tasks.*"]),
    # entry_points={
    #     "console_scripts": ["pipenv=pipenv:cli", "pipenv-resolver=pipenv.resolver:main"]
    # },
    package_data={
        # "": ["LICENSE", "NOTICES"],
        #     "pipenv.vendor.requests": ["*.pem"],
        #     "pipenv.vendor.certifi": ["*.pem"],
        #     "pipenv.vendor.click_completion": ["*.j2"],
        #     "pipenv.patched.notpip._vendor.certifi": ["*.pem"],
        #     "pipenv.patched.notpip._vendor.requests": ["*.pem"],
        #     "pipenv.patched.notpip._vendor.distlib._backport": ["sysconfig.cfg"],
        #     "pipenv.patched.notpip._vendor.distlib": [
        #         "t32.exe",
        #         "t64.exe",
        #         "w32.exe",
        #         "w64.exe",
        #     ],
    },
    python_requires=">=3.6",
    setup_requires=[],
    install_requires=required,
    extras_require={},
    include_package_data=True,
    license="Apache 2.0",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    cmdclass={"upload": UploadCommand, "deb": DebCommand},
)



import io
import json
import gzip

import graphene
import yaml
from requests.models import Request as RequestsRequest
from requests.structures import CaseInsensitiveDict
from werkzeug.wrappers import Request as WerkzeugRequest
from werkzeug.wrappers import BaseResponse as WerkzeugResponse


from urllib.parse import parse_qs

from .status import HTTP_200

# @staticmethod
# def funcname(parameter_list):
#     pass


class Request:
    def __init__(self):
        super().__init__()
        self._wz = None

    @classmethod
    def from_environ(kls, environ):
        self = kls()
        self._wz = WerkzeugRequest(environ)
        self.headers = CaseInsensitiveDict(self._wz.headers.to_list())
        self.method = self._wz.method
        self.full_url = self._wz.url
        self.url = self._wz.base_url
        self.full_path = self._wz.full_path
        self.path = self._wz.path
        self.params = parse_qs(self._wz.query_string.decode("utf-8"))
        self.raw = self._wz.stream
        self.content = self._wz.get_data(cache=True, as_text=False)
        self.text = self._wz.get_data(cache=True, as_text=True)
        self.data = self._wz.get_data(cache=True, as_text=True, parse_form_data=True)

        return self

    @property
    def accepts_yaml(self):
        return "yaml" in self.headers["accept"]

    @property
    def accepts_json(self):
        return "json" in self.headers["accept"]

    def json(self):
        return json.loads(self.content)

    def yaml(self):
        return yaml.load(self.content)


class Response:
    def __init__(self, req):
        self.req = req
        self.status_code = HTTP_200
        self.text = None
        self.content = None
        self.encoding = None
        self.media = None
        self.mimetype = None
        self.headers = {}

    @property
    def body(self):

        if self.content:
            return (self.content, self.mimetype, {})

        if self.text:
            return (
                self.text.encode("utf-8"),
                self.mimetype or "application/text",
                {"Encoding": "utf-8"},
            )

        if self.media:
            if self.req.accepts_yaml:
                return (
                    yaml.dump(self.media).encode("utf-8"),
                    self.mimetype or "application/x-yaml",
                    {},
                )
            if self.req.accepts_json:
                return (json.dumps(self.media), self.mimetype or "application/json", {})

    @property
    def gzipped_body(self):

        body, mimetype, headers = self.body
        # print(self.req.headers)
        if "gzip" in self.req.headers["Accept-Encoding"].lower():
            gzip_buffer = io.BytesIO()
            gzip_file = gzip.GzipFile(mode="wb", fileobj=gzip_buffer)
            gzip_file.write(body)
            gzip_file.close()

            new_headers = {
                "Content-Encoding": "gzip",
                "Vary": "Accept-Encoding",
                "Content-Length": str(len(body)),
            }
            headers.update(new_headers)
            # print(headers)

            return (gzip_buffer.getvalue(), mimetype, headers)
        else:
            return (body, mimetype, headers)

    @property
    def _wz(self):
        body, mimetype, headers = self.body
        headers.update(self.headers)

        r = WerkzeugResponse(
            body,
            status=self.status_code,
            mimetype=self.mimetype or mimetype,
            direct_passthrough=False,
        )
        r.headers = headers
        return r

    def __call__(self, environ, start_response):
        return self._wz(environ, start_response)


class Schema(graphene.Schema):
    def on_request(self, req, resp):
        pass
